# Generated by Django 4.0.3 on 2022-03-06 04:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_apscheduler", "0008_remove_djangojobexecution_started"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="djangojobexecution",
            constraint=models.UniqueConstraint(
                fields=("job_id", "run_time"), name="unique_job_executions"
            ),
        ),
    ]



import warnings
from datetime import datetime
from unittest import mock

import pytest
from apscheduler import events
from apscheduler.events import JobExecutionEvent, JobSubmissionEvent
from django import db
from django.utils import timezone

from django_apscheduler.jobstores import (
    DjangoJobStore,
    register_job,
    register_events,
)
from django_apscheduler.models import DjangoJob, DjangoJobExecution
from tests import conftest
from tests.conftest import DummyScheduler, dummy_job


class TestDjangoResultStoreMixin:
    def test_start_gets_scheduler_lock(self):
        store = DjangoJobStore()

        store.start(DummyScheduler(), "djangojobstore")
        assert store.lock is not None

    @pytest.mark.django_db
    def test_handle_submission_event_not_supported_raises_exception(self, jobstore):
        event = JobSubmissionEvent(
            events.EVENT_ALL, "test_job", jobstore, [timezone.now()]
        )

        with pytest.raises(NotImplementedError):
            jobstore.handle_submission_event(event)

    @pytest.mark.django_db
    @pytest.mark.parametrize(
        "event_code",
        [
            events.EVENT_JOB_SUBMITTED,
            events.EVENT_JOB_MAX_INSTANCES,
        ],
    )
    def test_handle_submission_event_creates_job_execution(
            self, event_code, jobstore, create_add_job
    ):
        job = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
        event = JobSubmissionEvent(event_code, job.id, jobstore, [timezone.now()])
        jobstore.handle_submission_event(event)

        assert DjangoJobExecution.objects.filter(job_id=event.job_id).exists()

    @pytest.mark.django_db(transaction=True)
    def test_handle_submission_event_for_job_that_no_longer_exists_does_not_raise_exception(
            self, jobstore
    ):
        event = JobSubmissionEvent(
            events.EVENT_JOB_SUBMITTED, "finished_job", jobstore, [timezone.now()]
        )
        jobstore.handle_submission_event(event)

        assert not DjangoJobExecution.objects.filter(job_id=event.job_id).exists()

    @pytest.mark.django_db
    def test_handle_execution_event_not_supported_raises_exception(self, jobstore):
        event = JobExecutionEvent(
            events.EVENT_ALL, "test_job", jobstore, timezone.now()
        )

        with pytest.raises(NotImplementedError):
            jobstore.handle_execution_event(event)

    @pytest.mark.django_db
    def test_handle_execution_event_creates_job_execution(
        self, jobstore, create_add_job
    ):
        job = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
        event = JobExecutionEvent(
            events.EVENT_JOB_EXECUTED, job.id, jobstore, timezone.now()
        )
        jobstore.handle_execution_event(event)

        assert DjangoJobExecution.objects.filter(job_id=event.job_id).exists()

    @pytest.mark.django_db(transaction=True)
    def test_handle_execution_event_for_job_that_no_longer_exists_does_not_raise_exception_regression_116(
        self, jobstore
    ):
        # Test for regression https://github.com/jcass77/django-apscheduler/issues/116
        event = JobExecutionEvent(
            events.EVENT_JOB_EXECUTED, "finished_job", jobstore, timezone.now()
        )
        jobstore.handle_execution_event(event)

        assert not DjangoJobExecution.objects.filter(job_id=event.job_id).exists()

    @pytest.mark.django_db
    def test_handle_error_event_not_supported_raises_exception(self, jobstore):
        event = JobExecutionEvent(
            events.EVENT_ALL, "test_job", jobstore, timezone.now()
        )

        with pytest.raises(NotImplementedError):
            jobstore.handle_error_event(event)

    @pytest.mark.django_db
    @pytest.mark.parametrize(
        "event_code",
        [
            events.EVENT_JOB_MISSED,
            events.EVENT_JOB_ERROR,
        ],
    )
    def test_handle_error_event_creates_job_execution(
        self, jobstore, create_add_job, event_code
    ):
        job = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
        event = JobExecutionEvent(event_code, job.id, jobstore, timezone.now())
        jobstore.handle_error_event(event)

        assert DjangoJobExecution.objects.filter(job_id=event.job_id).exists()

    @pytest.mark.django_db
    def test_handle_error_event_no_exception_sets_exception_text(
        self, jobstore, create_add_job
    ):
        job = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
        event = JobExecutionEvent(
            events.EVENT_JOB_ERROR, job.id, jobstore, timezone.now()
        )
        jobstore.handle_error_event(event)

        ex = DjangoJobExecution.objects.get(job_id=event.job_id)

        assert "raised an error!" in ex.exception

    @pytest.mark.django_db(transaction=True)
    def test_handle_error_event_for_job_that_no_longer_exists_does_not_raise_exception(
            self, jobstore
    ):
        event = JobExecutionEvent(
            events.EVENT_JOB_ERROR, "finished_job", jobstore, timezone.now()
        )
        jobstore.handle_error_event(event)

        assert not DjangoJobExecution.objects.filter(job_id=event.job_id).exists()

    @pytest.mark.django_db
    def test_register_event_listeners_registers_listeners(self, jobstore):
        jobstore.register_event_listeners()
        registered_event_codes = [event[1] for event in jobstore._scheduler._listeners]

        assert all(
            event_code in registered_event_codes
            for event_code in [
                events.EVENT_JOB_SUBMITTED | events.EVENT_JOB_MAX_INSTANCES,
                events.EVENT_JOB_EXECUTED,
                events.EVENT_JOB_ERROR | events.EVENT_JOB_MISSED,
            ]
        )


class TestDjangoJobStore:
    """
    We use the APScheduler tests to verify that DjangoJobStore implements the interface correctly.

    This test class should only contain tests that are specific to DjangoJobStore

    See 'test_apscheduler_jobstore.py' for details
    """

    @pytest.mark.django_db(transaction=True)
    def test_lookup_job_does_retry_on_db_operational_error(self, jobstore):
        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.get",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore.lookup_job("some job")

            assert close_mock.call_count == 1

    @pytest.mark.django_db(transaction=True)
    def test_get_due_jobs_does_retry_on_db_operational_error(self, jobstore):
        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.filter",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore.get_due_jobs(datetime(2016, 5, 3))

            assert close_mock.call_count == 1

    @pytest.mark.django_db(transaction=True)
    def test_get_next_run_time_does_retry_on_db_operational_error(self, jobstore):
        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.filter",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore.get_next_run_time()

            assert close_mock.call_count == 1

    @pytest.mark.django_db(transaction=True)
    def test_add_job_does_retry_on_db_operational_error(self, jobstore, create_job):
        job = create_job(
            func=dummy_job,
            trigger="date",
            trigger_args={"run_date": datetime(2016, 5, 3)},
            id="test",
        )

        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.create",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore.add_job(job)

            assert close_mock.call_count == 1

    @pytest.mark.django_db(transaction=True)
    def test_update_job_does_retry_on_db_operational_error(self, jobstore, create_job):
        job = create_job(
            func=dummy_job,
            trigger="date",
            trigger_args={"run_date": datetime(2016, 5, 3)},
            id="test",
        )

        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.select_for_update",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore.update_job(job)

            assert close_mock.call_count == 1

    @pytest.mark.django_db(transaction=True)
    def test_remove_job_does_retry_on_db_operational_error(self, jobstore):
        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.select_for_update",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore.remove_job("some job")

            assert close_mock.call_count == 1

    @pytest.mark.django_db(transaction=True)
    def test_remove_all_jobs_does_retry_on_db_operational_error(self, jobstore):
        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.all",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore.remove_all_jobs()

            assert close_mock.call_count == 1

    @pytest.mark.django_db(transaction=True)
    def test_get_jobs_does_retry_on_db_operational_error(self, jobstore):
        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.jobstores.DjangoJob.objects.filter",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    jobstore._get_jobs()

            assert close_mock.call_count == 1


@pytest.mark.django_db
def test_register_events_raises_deprecation_warning(scheduler, jobstore):

    with warnings.catch_warnings(record=True) as w:

        register_events(scheduler, jobstore)
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "deprecated" in str(w[-1].message)


@pytest.mark.django_db
def test_register_job(scheduler, jobstore):
    register_job(scheduler, "interval", seconds=1)(dummy_job)
    scheduler.start()

    assert DjangoJob.objects.count() == 1


@pytest.mark.django_db
def test_register_job_raises_deprecation_warning(scheduler, jobstore):

    with warnings.catch_warnings(record=True) as w:

        register_job(scheduler, "interval", seconds=1)(dummy_job)
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "deprecated" in str(w[-1].message)



from datetime import datetime

import pytest

from apscheduler.jobstores.base import JobLookupError, ConflictingIdError

from tests.conftest import dummy_job

pytestmark = pytest.mark.django_db  # Give all tests access to DB


# NOTE: These tests are copy / paste equivalents of APScheduler's own jobstore tests.
# See: https://github.com/agronholm/apscheduler/blob/master/tests/test_jobstores.py


def dummy_job2():
    pass


def dummy_job3():
    pass


class DummyClass:
    def dummy_method(self, a, b):
        return a + b

    @classmethod
    def dummy_classmethod(cls, a, b):
        return a + b


def test_add_instance_method_job(jobstore, create_add_job):
    instance = DummyClass()
    initial_job = create_add_job(
        jobstore, instance.dummy_method, kwargs={"a": 1, "b": 2}
    )
    job = jobstore.lookup_job(initial_job.id)
    assert job.func(*job.args, **job.kwargs) == 3


def test_add_class_method_job(jobstore, create_add_job):
    initial_job = create_add_job(
        jobstore, DummyClass.dummy_classmethod, kwargs={"a": 1, "b": 2}
    )
    job = jobstore.lookup_job(initial_job.id)
    assert job.func(*job.args, **job.kwargs) == 3


def test_lookup_job(jobstore, create_add_job):
    initial_job = create_add_job(jobstore)
    job = jobstore.lookup_job(initial_job.id)
    assert job == initial_job


def test_lookup_nonexistent_job(jobstore):
    assert jobstore.lookup_job("foo") is None


def test_get_all_jobs(jobstore, create_add_job):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2013, 8, 14))
    job3 = create_add_job(jobstore, dummy_job2, datetime(2013, 7, 11), paused=True)
    jobs = jobstore.get_all_jobs()
    assert jobs == [job2, job1, job3]


def test_get_pending_jobs(jobstore, create_add_job, timezone):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    job3 = create_add_job(jobstore, dummy_job3, datetime(2013, 8, 14))
    create_add_job(jobstore, dummy_job3, datetime(2013, 7, 11), paused=True)
    jobs = jobstore.get_due_jobs(timezone.localize(datetime(2014, 2, 27)))
    assert jobs == [job3, job2]

    jobs = jobstore.get_due_jobs(timezone.localize(datetime(2013, 8, 13)))
    assert jobs == []


def test_get_pending_jobs_subsecond_difference(jobstore, create_add_job, timezone):
    job1 = create_add_job(jobstore, dummy_job, datetime(2014, 7, 7, 0, 0, 0, 401))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 7, 7, 0, 0, 0, 402))
    job3 = create_add_job(jobstore, dummy_job3, datetime(2014, 7, 7, 0, 0, 0, 400))
    jobs = jobstore.get_due_jobs(timezone.localize(datetime(2014, 7, 7, 1)))
    assert jobs == [job3, job1, job2]


def test_get_next_run_time(jobstore, create_add_job, timezone):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    create_add_job(jobstore, dummy_job3, datetime(2013, 8, 14))
    create_add_job(jobstore, dummy_job3, datetime(2013, 7, 11), paused=True)
    assert jobstore.get_next_run_time() == timezone.localize(datetime(2013, 8, 14))


def test_add_job_conflicting_id(jobstore, create_add_job):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3), id="blah")
    pytest.raises(
        ConflictingIdError,
        create_add_job,
        jobstore,
        dummy_job2,
        datetime(2014, 2, 26),
        id="blah",
    )


def test_update_job(jobstore, create_add_job, timezone):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    replacement = create_add_job(
        None, dummy_job, datetime(2016, 5, 4), id=job1.id, max_instances=6
    )
    assert replacement.max_instances == 6
    jobstore.update_job(replacement)

    jobs = jobstore.get_all_jobs()
    assert len(jobs) == 2
    assert jobs[0].id == job2.id
    assert jobs[1].id == job1.id
    assert jobs[1].next_run_time == timezone.localize(datetime(2016, 5, 4))
    assert jobs[1].max_instances == 6


@pytest.mark.parametrize(
    "next_run_time", [datetime(2013, 8, 13), None], ids=["earlier", "null"]
)
def test_update_job_next_runtime(jobstore, create_add_job, next_run_time, timezone):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    job3 = create_add_job(jobstore, dummy_job3, datetime(2013, 8, 14))
    job1.next_run_time = timezone.localize(next_run_time) if next_run_time else None
    jobstore.update_job(job1)

    if next_run_time:
        assert jobstore.get_next_run_time() == job1.next_run_time
    else:
        assert jobstore.get_next_run_time() == job3.next_run_time


@pytest.mark.parametrize(
    "next_run_time", [datetime(2013, 8, 13), None], ids=["earlier", "null"]
)
@pytest.mark.parametrize("index", [0, 1, 2], ids=["first", "middle", "last"])
def test_update_job_clear_next_runtime(
    jobstore, create_add_job, next_run_time, timezone, index
):
    """
    Tests that update_job() maintains the proper ordering of the jobs,
    even when their next run times are initially the same.
    """
    jobs = [
        create_add_job(jobstore, dummy_job, datetime(2014, 2, 26), "job%d" % i)
        for i in range(3)
    ]
    jobs[index].next_run_time = (
        timezone.localize(next_run_time) if next_run_time else None
    )
    jobstore.update_job(jobs[index])
    due_date = timezone.localize(datetime(2014, 2, 27))
    due_jobs = jobstore.get_due_jobs(due_date)

    assert len(due_jobs) == (3 if next_run_time else 2)
    due_job_ids = [job.id for job in due_jobs]
    if next_run_time:
        if index == 0:
            assert due_job_ids == ["job0", "job1", "job2"]
        elif index == 1:
            assert due_job_ids == ["job1", "job0", "job2"]
        else:
            assert due_job_ids == ["job2", "job0", "job1"]
    else:
        if index == 0:
            assert due_job_ids == ["job1", "job2"]
        elif index == 1:
            assert due_job_ids == ["job0", "job2"]
        else:
            assert due_job_ids == ["job0", "job1"]


def test_update_job_nonexistent_job(jobstore, create_add_job):
    job = create_add_job(None, dummy_job, datetime(2016, 5, 3))
    pytest.raises(JobLookupError, jobstore.update_job, job)


def test_one_job_fails_to_load(jobstore, create_add_job, monkeypatch, timezone):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))
    create_add_job(jobstore, dummy_job3, datetime(2013, 8, 14))

    # Make the dummy_job2 function disappear
    monkeypatch.delitem(globals(), "dummy_job3")

    jobs = jobstore.get_all_jobs()
    assert jobs == [job2, job1]

    assert jobstore.get_next_run_time() == timezone.localize(datetime(2014, 2, 26))


def test_remove_job(jobstore, create_add_job):
    job1 = create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    job2 = create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))

    jobstore.remove_job(job1.id)
    jobs = jobstore.get_all_jobs()
    assert jobs == [job2]

    jobstore.remove_job(job2.id)
    jobs = jobstore.get_all_jobs()
    assert jobs == []


def test_remove_nonexistent_job(jobstore):
    pytest.raises(JobLookupError, jobstore.remove_job, "blah")


def test_remove_all_jobs(jobstore, create_add_job):
    create_add_job(jobstore, dummy_job, datetime(2016, 5, 3))
    create_add_job(jobstore, dummy_job2, datetime(2014, 2, 26))

    jobstore.remove_all_jobs()
    jobs = jobstore.get_all_jobs()
    assert jobs == []


def test_repr_jobstore(jobstore):
    assert (
        repr(jobstore)
        == f"<DjangoJobStore(pickle_protocol={jobstore.pickle_protocol})>"
    )



from datetime import timedelta
from unittest import mock

import pytest
from apscheduler.schedulers.background import BackgroundScheduler
from django.conf import settings
from django.contrib.messages.storage.base import BaseStorage
from django.utils import timezone
from django.utils.html import format_html

from django_apscheduler.admin import DjangoJobAdmin, DjangoJobExecutionAdmin
from django_apscheduler.jobstores import DjangoJobStore
from django_apscheduler.models import DjangoJob, DjangoJobExecution


class TestDjangoJobAdmin:
    @pytest.mark.django_db
    def test_get_queryset_calculates_averages(self, rf, request):
        now = timezone.now()
        run_time = now - timedelta(seconds=60)

        job = DjangoJob.objects.create(id="test_job", next_run_time=run_time)
        request.addfinalizer(job.delete)

        DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SUCCESS,
            run_time=run_time - timedelta(seconds=60),
            duration=5,
            finished=(run_time + timedelta(seconds=5)).timestamp(),
        )  # Old job execution

        DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SUCCESS,
            run_time=run_time,
            duration=10,
            finished=(run_time + timedelta(seconds=10)).timestamp(),
        )  # Most recent job execution

        admin = DjangoJobAdmin(DjangoJob, None)
        admin.get_queryset(rf.get("/admin/django_apscheduler/djangojob"))

        assert admin.avg_duration_qs.count() == 1
        assert admin.avg_duration_qs.first()[0] == job.id
        assert admin.avg_duration_qs.first()[1] == 7.5

    @pytest.mark.django_db
    def test_local_run_time_returns_paused_if_no_run_time_scheduled(self, rf, request):
        job = DjangoJob.objects.create(id="test_job")
        request.addfinalizer(job.delete)

        admin = DjangoJobAdmin(DjangoJob, None)
        admin.get_queryset(rf.get("/admin/django_apscheduler/djangojob"))

        assert admin.local_run_time(job) == "(paused)"

    @pytest.mark.django_db
    def test_average_duration_returns_correct_value(self, rf, request):
        now = timezone.now()
        run_time = now - timedelta(seconds=60)

        job = DjangoJob.objects.create(id="test_job", next_run_time=run_time)
        request.addfinalizer(job.delete)

        DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SUCCESS,
            run_time=run_time - timedelta(seconds=60),
            duration=5,
            finished=(run_time + timedelta(seconds=5)).timestamp(),
        )  # Old job execution

        DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SUCCESS,
            run_time=run_time,
            duration=10,
            finished=(run_time + timedelta(seconds=10)).timestamp(),
        )  # Most recent job execution

        admin = DjangoJobAdmin(DjangoJob, None)
        admin.get_queryset(rf.get("/admin/django_apscheduler/djangojob"))

        assert admin.average_duration(job) == 7.5

    @pytest.mark.django_db
    def test_average_duration_no_executions_shows_none_text(self, request, rf):
        now = timezone.now()
        run_time = now - timedelta(seconds=60)

        job = DjangoJob.objects.create(id="test_job", next_run_time=run_time)
        request.addfinalizer(job.delete)

        admin = DjangoJobAdmin(DjangoJob, None)
        r = rf.get("/django_apscheduler/djangojob/")
        admin.get_queryset(r)

        assert admin.average_duration(job) == "None"

    @pytest.mark.django_db(transaction=True)
    def test_run_selected_jobs_creates_job_execution_entry(self, rf, monkeypatch):
        monkeypatch.setattr(
            settings, "APSCHEDULER_RUN_NOW_TIMEOUT", 1
        )  # Shorten timeout to reduce test runtime

        scheduler = BackgroundScheduler()
        scheduler.add_jobstore(DjangoJobStore())
        scheduler.start()

        job = scheduler.add_job(print, trigger="interval", seconds=60)

        admin = DjangoJobAdmin(DjangoJob, None)

        r = rf.get("/django_apscheduler/djangojob/")
        # Add support for Django messaging framework
        r._messages = mock.MagicMock(BaseStorage)
        r._messages.add = mock.MagicMock()

        assert not DjangoJobExecution.objects.filter(job_id=job.id).exists()

        admin.run_selected_jobs(r, DjangoJob.objects.filter(id=job.id))

        assert DjangoJobExecution.objects.filter(job_id=job.id).exists()
        r._messages.add.assert_called_with(20, f"Executed job '{job.id}'!", "")

        scheduler.shutdown()

    @pytest.mark.django_db(transaction=True)
    def test_run_selected_jobs_job_not_found_skips_execution(self, rf):
        scheduler = BackgroundScheduler()
        scheduler.add_jobstore(DjangoJobStore())
        scheduler.start()

        job = DjangoJob.objects.create(id="test_job")

        admin = DjangoJobAdmin(DjangoJob, None)

        r = rf.get("/django_apscheduler/djangojob/")
        # Add support for Django messaging framework
        r._messages = mock.MagicMock(BaseStorage)
        r._messages.add = mock.MagicMock()

        admin.run_selected_jobs(r, DjangoJob.objects.filter(id=job.id))

        assert DjangoJobExecution.objects.count() == 0
        r._messages.add.assert_called_with(
            30, "Could not find job test_job in the database! Skipping execution...", ""
        )

        scheduler.shutdown()

    @pytest.mark.django_db(transaction=True)
    def test_run_selected_jobs_enforces_timeout(self, rf, monkeypatch):
        monkeypatch.setattr(
            settings, "APSCHEDULER_RUN_NOW_TIMEOUT", 1
        )  # Shorten timeout to reduce test runtime

        scheduler = BackgroundScheduler()
        scheduler.add_jobstore(DjangoJobStore())
        scheduler.start()

        job = scheduler.add_job(print, trigger="interval", seconds=60)

        admin = DjangoJobAdmin(DjangoJob, None)

        r = rf.get("/django_apscheduler/djangojob/")
        # Add support for Django messaging framework
        r._messages = mock.MagicMock(BaseStorage)
        r._messages.add = mock.MagicMock()

        with mock.patch("django_apscheduler.admin.BackgroundScheduler.add_listener"):
            admin.run_selected_jobs(r, DjangoJob.objects.filter(id=job.id))

        assert DjangoJobExecution.objects.count() == 0
        r._messages.add.assert_called_with(
            40,
            format_html(
                "Maximum runtime of {} seconds exceeded! Not all jobs could be completed successfully. "
                "Pending jobs: {}",
                admin._job_execution_timeout,
                ",".join({job.id}),
            ),
            "",
        )

        scheduler.shutdown()


class TestDjangoJobExecutionAdmin:
    @pytest.mark.django_db
    def test_html_status_returns_colored_status_text(self, rf, request):
        now = timezone.now()

        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        execution = DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SUCCESS,
            run_time=now,
        )

        admin = DjangoJobExecutionAdmin(DjangoJob, None)
        admin.get_queryset(rf.get("/admin/django_apscheduler/djangojob"))

        assert "green" in admin.html_status(execution)

    @pytest.mark.django_db
    def test_duration_text_no_duration_returns_na(self, rf, request):
        now = timezone.now()

        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        execution = DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SUCCESS,
            run_time=now,
        )

        admin = DjangoJobExecutionAdmin(DjangoJob, None)
        admin.get_queryset(rf.get("/admin/django_apscheduler/djangojob"))

        assert admin.duration_text(execution) == "N/A"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
-------------------------------------------------
   File Name：      register
   Description:
   Author:          dingyong.cui
   date：           2023/3/15
-------------------------------------------------
   Change Activity:
                    2023/3/15
-------------------------------------------------
"""
import importlib
import inspect
import logging
import pkgutil
from typing import List

logger = logging.getLogger(__name__)

_DEFAULT_RELATE_NAME = 'service'


class Register:

    def __init__(self, *args, **kwargs):
        super(Register, self).__init__()
        self._dict = {}

    def register(self, target):
        def add_register_item(key, value):
            if not callable(value):
                raise Exception(f"register object must be callable! But receive:{value} is not callable!")
            if key in self._dict:
                logger.warning(f"{value.__name__} has been registered before, so we will overriden it")
            self._dict[key] = value

            return value

        if callable(target):  # 如果传入的目标可调用，说明之前没有给出注册名字，我们就以传入的函数或者类的名字作为注册名
            return add_register_item(target.__name__, target)

        return lambda x: add_register_item(target, x)

    def __call__(self, target):
        return self.register(target)

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def __str__(self):
        return str(self._dict)

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()


aps_register = Register()


def autodiscover_aps(relate_name=None) -> List:
    aps_funcs = []
    if relate_name is None:
        relate_name = _DEFAULT_RELATE_NAME

    from django.apps import apps
    for app_config in apps.get_app_configs():
        pkg = app_config.module
        pkg_path = f'{pkg.__path__[0]}\\{relate_name}'
        pkg_name = f'{pkg.__name__}.{relate_name}'
        part_aps_funcs = _autodiscover_aps(pkg_path, pkg_name)
        aps_funcs.extend(part_aps_funcs)

    return aps_funcs


def _autodiscover_aps(pkg_path, pkg_name):
    part_aps_funcs = []
    for _, file, is_pkg in pkgutil.iter_modules([pkg_path]):
        py_module = importlib.import_module(f'.{file}', package=pkg_name)
        for name, p_cls in inspect.getmembers(py_module):
            if isinstance(p_cls, Register):
                for func in p_cls.items():
                    aps_func = _convert_func(func)
                    part_aps_funcs.append(aps_func)

    return part_aps_funcs


def _convert_func(func: tuple):
    func = func[1]
    func_module = func.__module__
    func_name = func.__qualname__
    func_params = inspect.getfullargspec(func).args
    func_params = [fp for fp in func_params if fp != 'self']

    aps_func = {
        'func_module': func_module,
        'func': func_name,
        'func_params': func_params
    }

    return aps_func



from datetime import datetime
from unittest import mock

import pytest
import pytz
from django import db

from django.utils import timezone

from django_apscheduler import util


def test_get_dt_format_default():
    assert util.get_dt_format() == "N j, Y, f:s a"


def test_get_dt_format_gets_format_from_settings(use_seconds_format):
    assert util.get_dt_format() == use_seconds_format


def test_get_local_dt_format_naive(use_hour_format):

    dt = datetime.utcnow()
    dt_hour = dt.strftime("%H")

    localized_dt_hour = util.get_local_dt_format(dt)

    assert localized_dt_hour == dt_hour


def test_get_local_dt_format_aware(use_tz):
    utc_dt = timezone.now()

    local_dt = timezone.localtime(utc_dt)
    local_dt_hour = local_dt.strftime("%H")

    localized_dt_hour = util.get_local_dt_format(utc_dt)

    assert localized_dt_hour == local_dt_hour


def test_get_django_internal_datetime_makes_naive_if_django_timzone_support_disabled(
    settings,
):
    settings.USE_TZ = False
    internal_dt = util.get_django_internal_datetime(datetime.now(tz=pytz.utc))

    assert timezone.is_naive(internal_dt)


def test_get_django_internal_datetime_makes_aware_if_django_timezone_support_enabled(
    settings,
):
    settings.USE_TZ = True
    internal_dt = util.get_django_internal_datetime(datetime.now())

    assert timezone.is_aware(internal_dt)


@pytest.mark.django_db
def test_get_apscheduler_datetime(scheduler):
    apscheduler_dt = util.get_apscheduler_datetime(datetime.now(), scheduler)

    assert timezone.is_aware(apscheduler_dt)


@pytest.mark.django_db
def test_retry_on_db_operational_error_no_db_errors(caplog):
    @util.retry_on_db_operational_error
    def dummy_db_op():
        return

    with mock.patch.object(db.connection, "close") as close_mock:
        dummy_db_op()
        assert not close_mock.called

    assert "Retrying with a new DB connection..." not in caplog.text


@pytest.mark.django_db
def test_retry_on_db_operational_error_db_operational_error_retry_ok(caplog):
    def dummy_func_maker():
        call_count = 0

        @util.retry_on_db_operational_error
        def dummy_db_op():
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                # Raise exception on first call
                raise db.OperationalError("Some DB-related error")

            return call_count

        return dummy_db_op

    func = dummy_func_maker()
    with mock.patch.object(db.connection, "close") as close_mock:
        call_count = func()
        assert call_count == 2
        assert close_mock.call_count == 1

    assert (
            "DB error executing 'dummy_db_op' (Some DB-related error). Retrying with a new DB connection..."
            in caplog.text
    )


@pytest.mark.django_db
def test_retry_on_db_operational_error_db_operational_error_retry_error_persists_re_raises(
        caplog,
):
    @util.retry_on_db_operational_error
    def func():
        raise db.OperationalError("Some DB-related error")

    with mock.patch.object(db.connection, "close") as close_mock:
        with pytest.raises(db.OperationalError, match="Some DB-related error"):
            call_count = func()
            assert call_count == 2

        assert close_mock.call_count == 1

    assert (
            "DB error executing 'func' (Some DB-related error). Retrying with a new DB connection..."
            in caplog.text
    )


@pytest.mark.django_db
def test_retry_on_db_operational_error_non_db_operational_error_re_raises(
        caplog,
):
    @util.retry_on_db_operational_error
    def func():
        raise RuntimeError("Some non DB-related error")

    with mock.patch.object(db.connection, "close") as close_mock:
        with pytest.raises(RuntimeError, match="Some non DB-related error"):
            func()

        assert not close_mock.called

    assert (
            "DB error executing 'dummy_db_op' (Some DB-related error). Retrying with a new DB connection..."
            not in caplog.text
    )


def test_close_old_connections_calls_close_old_connections():
    @util.close_old_connections
    def job_mock():
        pass

    with mock.patch(
            "django_apscheduler.util.db.close_old_connections"
    ) as close_old_connections_mock:
        job_mock()

    assert close_old_connections_mock.call_count == 2


def test_close_old_connections_even_if_exception_is_raised():
    @util.close_old_connections
    def job_mock():
        raise RuntimeError("some error")

    with mock.patch(
            "django_apscheduler.util.db.close_old_connections"
    ) as close_old_connections_mock:
        with pytest.raises(RuntimeError, match="some error"):
            job_mock()

    assert close_old_connections_mock.call_count == 2



import logging
from datetime import datetime
from functools import wraps

from apscheduler.schedulers.base import BaseScheduler
from django import db
from django.conf import settings
from django.utils import formats
from django.utils import timezone

logger = logging.getLogger(__name__)


def get_dt_format() -> str:
    """Return the configured format for displaying datetimes in the Django admin views"""
    return formats.get_format(
        getattr(settings, "APSCHEDULER_DATETIME_FORMAT", "N j, Y, f:s a")
    )


def get_local_dt_format(dt: datetime) -> str:
    """Get the datetime in the localized datetime format"""
    if dt and settings.USE_TZ and timezone.is_aware(dt):
        dt = timezone.localtime(dt)

    return formats.date_format(dt, get_dt_format())


def get_django_internal_datetime(dt: datetime) -> datetime:
    """
    Get the naive or aware version of the datetime based on the configured `USE_TZ` Django setting. This is also the
    format that Django uses to store datetimes internally.
    """
    if dt:
        if settings.USE_TZ and timezone.is_naive(dt):
            return timezone.make_aware(dt)

        elif not settings.USE_TZ and timezone.is_aware(dt):
            return timezone.make_naive(dt)

    return dt


def get_apscheduler_datetime(dt: datetime, scheduler: BaseScheduler) -> datetime:
    """
    Make the datetime timezone aware (if necessary), using the same timezone as is currently configured for the
    scheduler.
    """
    if dt and timezone.is_naive(dt):
        return timezone.make_aware(dt, timezone=scheduler.timezone)

    return dt


def retry_on_db_operational_error(func):
    """
    This decorator can be used to wrap a database-related method so that it will be retried when a
    django.db.OperationalError or django.db.InterfaceError is encountered.

    The rationale is that these exceptions are usually raised when attempting to use an old database connection that
    the database backend has since closed. Closing the Django connection as well, and re-trying with a fresh connection,
    is usually sufficient to solve the problem.

    It is a reluctant workaround for users that persistently have issues with stale database connections (most notably:
    2006, 'MySQL server has gone away').

    The recommended approach is still to rather install a database connection pooler (like pgbouncer), to take care of
    database connection management for you, but the issue has been raised enough times by different individuals that a
    workaround is probably justified.

    CAUTION: any method that this decorator is applied to MUST be idempotent (i.e. the method can be retried a second
    time without any unwanted side effects). If your method performs any actions before the database exception is
    raised then those actions will be repeated. If you don't want that to happen then it would be best to handle the
    exception manually and call `db.close_old_connections()` in an appropriate fashion inside your own method instead.

    The following list of alternative workarounds were also considered:

    1. Calling db.close_old_connections() pre-emptively before the job store executes a DB operation: this would break
       Django's standard connection management. For example, if the `CONN_MAX_AGE` setting is set to 0, a new connection
       will be required for *every* database operation (as opposed to at the end of every *request* like in the Django
       standard). The database overhead, and associated performance penalty, that this approach would impose seem
       unreasonable. See: https://docs.djangoproject.com/en/dev/ref/settings/#std:setting-CONN_MAX_AGE.

    2. Using a custom QuerySet or database backend that handles the relevant database exceptions automatically: this
       would be more convenient than having to decorate individual methods, but it would also break when a DB operation
       needs to be re-tried as part of an atomic transaction. See: https://github.com/django/django/pull/2740

    3. Pinging the database before each operation to see if it is still available: django-apscheduler used to make use
       of this approach (see: https://github.com/jcass77/django-apscheduler/blob/9ac06b33d19961da6c36d5ac814d4338beb11309/django_apscheduler/models.py#L16-L51).
       Injecting an additional database query, on an arbitrary schedule, seems like an unreasonable thing to do,
       especially considering that this would be unnecessary for users that already make use of a database connection
       pooler to manage their connections properly.
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except (db.OperationalError, db.InterfaceError) as e:
            logger.warning(
                f"DB error executing '{func.__name__}' ({e}). Retrying with a new DB connection..."
            )
            db.close_old_connections()
            result = func(*args, **kwargs)

        return result

    return func_wrapper


def close_old_connections(func):
    """
    A decorator that ensures that Django database connections that have become unusable, or are obsolete, are closed
    before and after a method is executed (see: https://docs.djangoproject.com/en/dev/ref/databases/#general-notes
    for background).

    This decorator is intended to be used to wrap APScheduler jobs, and provides functionality comparable to the
    Django standard approach of closing old connections before and after each HTTP request is processed.

    It only makes sense for APScheduler jobs that require database access, and prevents `django.db.OperationalError`s.
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        db.close_old_connections()
        try:
            result = func(*args, **kwargs)
        finally:
            db.close_old_connections()

        return result

    return func_wrapper



import logging
from datetime import timedelta
from threading import RLock
from unittest import mock

import pytest
from apscheduler import events
from django import db
from django.utils import timezone

from django_apscheduler.models import DjangoJobExecution, DjangoJob
from tests import conftest

logging.basicConfig()


class TestDjangoJob:
    @pytest.mark.django_db
    def test_str(self, request):
        job = DjangoJob.objects.create(id="test_job", next_run_time=timezone.now())
        request.addfinalizer(job.delete)

        assert str(job).startswith("test_job (next run at:")

    @pytest.mark.django_db
    def test_str_paused(self, request):
        job = DjangoJob.objects.create(id="test_job")
        request.addfinalizer(job.delete)

        assert str(job) == "test_job (paused)"


class TestDjangoJobExecutionManager:
    @pytest.mark.django_db
    def test_delete_old_job_executions_deletes_old_jobs(
        self, request, jobstore, scheduler
    ):
        now = timezone.now()
        first_run = now - timedelta(seconds=10)
        second_run = now - timedelta(seconds=5)
        duration = 5.0

        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        DjangoJobExecution.objects.create(
            job=job,
            status=events.EVENT_JOB_EXECUTED,
            run_time=first_run,
            duration=duration,
            finished=(first_run + timedelta(seconds=duration)).timestamp(),
        )  # Old execution

        DjangoJobExecution.objects.create(
            job=job,
            status=events.EVENT_JOB_EXECUTED,
            run_time=second_run,
            duration=duration,
            finished=(first_run + timedelta(seconds=duration)).timestamp(),
        )  # Recent execution

        assert DjangoJobExecution.objects.count() == 2

        DjangoJobExecution.objects.delete_old_job_executions(duration + 1)

        assert DjangoJobExecution.objects.count() == 1


class TestDjangoJobExecution:
    @pytest.mark.django_db
    def test_atomic_update_or_create_creates_new_job(self, request, jobstore):
        now = timezone.now()
        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        DjangoJobExecution.atomic_update_or_create(
            RLock(),
            job.id,
            job.next_run_time - timedelta(seconds=5),
            DjangoJobExecution.SUCCESS,
        )

        assert DjangoJobExecution.objects.filter(job_id=job.id).exists()

    @pytest.mark.django_db
    def test_atomic_update_or_create_updates_existing_jobs(self, request, jobstore):
        now = timezone.now()
        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        ex = DjangoJobExecution.objects.create(
            job_id=job.id,
            run_time=job.next_run_time - timedelta(seconds=5),
            status=DjangoJobExecution.SENT,
        )

        assert ex.status == DjangoJobExecution.SENT
        assert ex.duration is None
        assert ex.finished is None

        DjangoJobExecution.atomic_update_or_create(
            RLock(),
            ex.job_id,
            ex.run_time,
            DjangoJobExecution.SUCCESS,
        )

        ex.refresh_from_db()

        assert ex.status == DjangoJobExecution.SUCCESS
        assert ex.duration is not None
        assert ex.finished is not None

    @pytest.mark.django_db
    def test_atomic_update_or_create_ignores_late_submission_events(
        self, request, jobstore
    ):
        now = timezone.now()
        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        ex = DjangoJobExecution.objects.create(
            job_id=job.id,
            run_time=job.next_run_time - timedelta(seconds=5),
            status=DjangoJobExecution.SUCCESS,
        )

        assert ex.status == DjangoJobExecution.SUCCESS
        assert ex.duration is None
        assert ex.finished is None

        DjangoJobExecution.atomic_update_or_create(
            RLock(),
            ex.job_id,
            ex.run_time,
            DjangoJobExecution.SENT,
        )

        ex.refresh_from_db()

        assert ex.status == DjangoJobExecution.SUCCESS
        assert ex.duration is None
        assert ex.finished is None

    @pytest.mark.django_db(transaction=True)
    def test_atomic_update_or_create_does_retry_on_db_operational_error(
            self, request, jobstore
    ):
        now = timezone.now()
        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        ex = DjangoJobExecution.objects.create(
            job_id=job.id,
            run_time=job.next_run_time - timedelta(seconds=5),
            status=DjangoJobExecution.SENT,
        )

        with mock.patch.object(db.connection, "close") as close_mock:
            with pytest.raises(db.OperationalError, match="Some DB-related error"):
                with mock.patch(
                        "django_apscheduler.models.DjangoJobExecution.objects.select_for_update",
                        side_effect=conftest.raise_db_operational_error,
                ):
                    DjangoJobExecution.atomic_update_or_create(
                        RLock(),
                        ex.job_id,
                        ex.run_time,
                        DjangoJobExecution.SUCCESS,
                    )

            assert close_mock.call_count == 1

    @pytest.mark.django_db
    def test_str(self, request):
        now = timezone.now()
        duration = 1
        prev_run = now - timedelta(seconds=5)

        job = DjangoJob.objects.create(id="test_job", next_run_time=now)
        request.addfinalizer(job.delete)

        ex = DjangoJobExecution.objects.create(
            job=job,
            status=DjangoJobExecution.SUCCESS,
            run_time=prev_run,
            duration=duration,
            finished=(prev_run + timedelta(duration)).timestamp(),
        )

        assert str(ex) == f"{ex.id}: job '{job.id}' ({DjangoJobExecution.SUCCESS})"
